#!/bin/bash

bailout() {
    echo "$@" >&2
    exit 3
}

build1() {
    #
    # Process one skeleton
    #
    local srcpath dstpath
    srcpath=$1
    dstpath=${srcpath%.skel}
    case $dstpath in
        *.md)   cat $srcpath \
                  | expand_includes \
                  | expand_variables \
                > $dstpath ;;
        *)      cat $srcpath \
                  | expand_variables \
                > $dstpath ;;
    esac
    echo $dstpath >> built.list
}

build() {
    #
    # Add meat to all skeletons
    #
    local srcpath
    find -type f -name '*.skel' \
     | while read srcpath;
       do
           build1 "$srcpath"
       done
}

clean() {
    #
    # Clean up tree after building
    #
    test -f built.list && {
        cat built.list | xargs rm -f
        rm -f built.list
    } || :
}

configure() {
    #
    # Load variables fdom config.mk
    #
    tmp=$(mktemp)
    sed -e 's/ = /=/' < config.mk > $tmp
    . $tmp
    rm -f $tmp
    PREFIX="$(readlink -f "$PREFIX")"
}

die() {
    #
    # Exit with message and non-zero exit status
    #
    echo "$@" >&2
    exit 4
}

dist() {
    #
    # Create distributable tarball
    #
    local version=$(get_version)
    local dirname=ffoo-$version
    mkdir -p $dirname
    local item
    cp -R $(list_of_dist_items) $dirname
    sed -i -e "s/^VERSION = .*/VERSION = $version/" $dirname/config.mk
    tar -cf $dirname.tar $dirname
    gzip $dirname.tar
    rm -rf $dirname
}

expand_includes() {
    #
    # Expand include directives
    #
    # Expand e.g. `<!-- include4: foo.sh -->` to include code of foo.sh
    #
    perl -we '
        use strict;
        my $text;
        while (<>) {
            chomp;
            if (m/<!-- include4: (\S+) -->/) {
                open my $fh, $1 or warn "cannot find: $1";
                my $text = do { local($/); <$fh> };
                close $fh;
                $text =~ s/^(.)/    $1/gm;
                chomp $text;
                print "$text\n";
            } else {
                print "$_\n";
            }
        }
    '
}

expand_groups() {
    #
    # Read path groups from paths.ini and expand to install params
    #
    local group
    local mode
    cat "mkit/paths.ini" \
      | read_section groups \
      | while read line;
        do
            group=${line%=*}
            mode=${line#$group=}
            expand_paths "$group" "$mode"
        done
}

expand_paths() {
    #
    # For each path of group $1, print all install params
    #
    local group=$1
    local mode=$2
    local root="$(get_root $group)"
    local src dst
    cat mkit/paths.ini \
      | read_section $group \
      | while read line;
        do
            src=${line%=*}
            dst=${line#$src=}
            echo "$src:$mode:$root/$dst"
        done
}

expand_variables() {
    #
    # Expand variable values
    #
    perl -pe "
        s|__FFOO_DIR__|$SHARE_DIR/ffoo|;
        s|__CODENAME__|$CODENAME|;
        s|__VERSION__|$(get_version)|;
    "
}

get_root() {
    #
    # Find out target rooot for group $1
    #
    local group="$1"
    local varname="${group^^*}_DIR"
    test -n "${!varname}" || die "missing in config.mk: $varname"
    echo "${!varname}"
}

get_version() {
    #
    # Build semver version string with build metadata
    #
    # Build version string from available info using following
    # logic:
    #
    #  1. use VERSION (from config.mk)
    #  2. if we are in git, override the version with last tag
    #  3. if set, add STAGE (from config.mk) as pre-release ID
    #     (afer dash)
    #  4. if we are at a later commit than the last tag, add commit
    #     sha1 to build metadata (after plus sign)
    #  5. if the tree is "dirty", i.e. has uncommited changes,
    #     add "dirty" to build metadata
    #
    # The version is compatible with SemVer 2.0.0.
    #
    # Examples:
    #
    #     myprog v1.0.7                         # all clear
    #     myprog v1.0.7-alpha                   # STAGE="alpha"
    #     myprog v1.0.7-alpha+g1aef811          # ^^ + some commits after
    #     myprog v1.0.7-alpha+g1aef811.dirty    # ^^ + tree edited
    #     myprog v1.0.7-alpha+dirty             # tag OK but tree edited
    #     myprog v1.0.7+dirty                   # ^^ but no stage
    #
    # Note that versions with "dirty" should be perceived as kind of
    # dangerous outside developer's own machine.  Versions with sha1 are
    # safer but must not be released.
    #
    # I have considered decorating the git commit refs to make them
    # sort of sortable (e.g. "r1.g1aef811"), but on second thought,
    # I don't think it's good idea to give *any* semantics to meta-data
    # at all.  First, there is no rule that r1<r2<r3; a commit can be
    # removing what other just added and one change can be split to
    # multiple commits.  Also, the whole thing breaks anyway once you
    # rebase your branch (no, it's not a sin).  The sole purpose of
    # meta-data is to *identify* the code, and provide safe path back
    # to tree; commit refs are already perfect for that.
    #
    # FIXME:  The assumpton "if we are in git repo, we can read version
    #         from tags" fails if we are in shallow clone made from
    #         other than a tagged commit.
    #
    local version=$VERSION
    local stage=$STAGE
    if git rev-parse HEAD >&/dev/null;
    then    # we are in git repo... so we can get smart
        local lasttag=$(git tag | grep ^v | sort -V | tail -n1)
        if ! git describe --tags --exact-match HEAD >&/dev/null;
        then    # we are not at later commit than the last tag
            local sha=g$(git log -1 --pretty=format:%h HEAD)
        fi
        if test "$(git diff --shortstat 2>/dev/null)" != "";
        then    # thr tree is "dirty", i.e. has been edited
            local dirty=dirty
        fi
        version=${lasttag:1}
        local suffix=""
        case $stage:$sha:$dirty in
            ::)         suffix=""                    ;;
            ::dirty)    suffix="+$dirty"             ;;
            :g*:)       suffix="+$sha"               ;;
            :g*:dirty)  suffix="+$sha.dirty"         ;;
            *::)        suffix="-$stage"             ;;
            *::dirty)   suffix="-$stage+$dirty"      ;;
            *:g*:)      suffix="-$stage+$sha"        ;;
            *:g*:dirty) suffix="-$stage+$sha.$dirty" ;;
        esac
        version=$version$suffix
    fi
    echo $version
}

install() {
    #
    # Install product
    #
    expand_groups \
      | while read line;
        do
            src="$(cut -d: -f1 <<<"$line")"
            mode="$(cut -d: -f2 <<<"$line")"
            dst="$(cut -d: -f3- <<<"$line")"
            command -p install -DTvm $mode $src "$dst"
        done
    test -f .autoclean && clean || :
}

read_section() {
    #
    # Read one INI section
    #
    local wnt=$1
    local ok=false
    grep '.' \
      | grep -v '\s*#' \
      | while read line;
        do
            case "$line" in
                \[$wnt\]) ok=true;  continue ;;
                \[*\])    ok=false; continue ;;
            esac
            $ok || continue
            printf -- "%s\n" "$line"
        done \
      | sed -e 's/ *= */=/; s/ +$//; s/^//;'
}

route() {
    #
    # Call correct function based on $1
    #
    case $1 in
        build|clean|dist|install|install_manpages|manpages|uninstall)
            $1
            ;;
        *)
            echo "usage: $(basename $0) build|clean|dist|install|uninstall" >&2
    esac
}

uninstall() {
    #
    # Uninstall product
    #
    expand_groups \
      | while read line;
        do
            src="$(cut -d: -f1 <<<"$line")"
            mode="$(cut -d: -f2 <<<"$line")"
            dst="$(cut -d: -f3- <<<"$line")"
            rm -vf "$dst"
        done

}
